# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing  # noqa
from urllib import parse

import netaddr

import tobiko
from tobiko.shell import sh
from tobiko.shell import ssh


CURL_CONNECTION_ERRORS = {
    7,   # Connection refused
    22,  # 404 Not Found
    28,  # Connection timedout
}


def execute_curl(
        hostname: typing.Union[str, netaddr.IPAddress] = None,
        port: int = None,
        path: str = None,
        scheme: str = None,
        ssh_client: ssh.SSHClientType = None,
        connect_timeout: tobiko.Seconds = None,
        fail_silently: bool = True,
        retry_count: int = None,
        retry_timeout: tobiko.Seconds = None,
        retry_interval: tobiko.Seconds = None,
        **execute_params) -> str:
    """Execute curl command

    Returns the command output in case of success,
    raises ShellCommandFailed otherwise.
    """
    netloc = make_netloc(hostname=hostname, port=port)
    url = make_url(scheme=scheme, netloc=netloc, path=path)
    command = sh.shell_command('curl -g')
    if fail_silently:
        command += '-f'
    if connect_timeout is not None:
        command += f'--connect-timeout {int(connect_timeout)}'
    command += url
    for attempt in tobiko.retry(count=retry_count,
                                timeout=retry_timeout,
                                interval=retry_interval,
                                default_count=1,
                                default_timeout=60.,
                                default_interval=1.):
        try:
            return sh.execute(command, ssh_client=ssh_client,
                              **execute_params).stdout
        except sh.ShellCommandFailed as ex:
            if ex.exit_status in CURL_CONNECTION_ERRORS:
                # Retry on connection errors
                attempt.check_limits()
            else:
                raise
    raise RuntimeError('Unexpected failure')


def make_netloc(
        hostname: typing.Union[str, netaddr.IPAddress, None] = None,
        port: typing.Optional[int] = None,
        username: typing.Optional[str] = None,
        password: typing.Optional[str] = None) -> str:
    if not hostname:
        return ''
    try:
        ip_address = netaddr.IPAddress(hostname)
    except netaddr.AddrFormatError:
        tobiko.check_valid_type(hostname, str)
        netloc = hostname
    else:
        if ip_address.version == 6:
            # Add square brackets around IPv6 address to please curl
            netloc = f'[{hostname}]'
        else:
            netloc = str(hostname)
    if port is not None:
        netloc = f'{netloc}:{port}'
    if username is not None:
        if password is not None:
            netloc = f'{username}:{password}@{netloc}'
        else:
            netloc = f'{username}@{netloc}'
    return netloc


def make_url(scheme: typing.Optional[str] = None,
             netloc: typing.Optional[str] = None,
             path: typing.Optional[str] = None) -> str:
    return parse.SplitResult(scheme=(scheme or 'http').lower(),
                             netloc=netloc or '',
                             path=path or '',
                             query='',
                             fragment='').geturl()



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.shell.sh import _cmdline
from tobiko.shell.sh import _command
from tobiko.shell.sh import _connection
from tobiko.shell.sh import _exception
from tobiko.shell.sh import _execute
from tobiko.shell.sh import _hostname
from tobiko.shell.sh import _io
from tobiko.shell.sh import _local
from tobiko.shell.sh import _nameservers
from tobiko.shell.sh import _path
from tobiko.shell.sh import _process
from tobiko.shell.sh import _ps
from tobiko.shell.sh import _reboot
from tobiko.shell.sh import _ssh
from tobiko.shell.sh import _systemctl
from tobiko.shell.sh import _uptime
from tobiko.shell.sh import _wc
from tobiko.shell.sh import _which


get_command_line = _cmdline.get_command_line

ShellCommand = _command.ShellCommand
ShellCommandType = _command.ShellCommandType
shell_command = _command.shell_command

LocalShellConnection = _connection.LocalShellConnection
ShellConnection = _connection.ShellConnection
ShellConnectionType = _connection.ShellConnectionType
SSHShellConnection = _connection.SSHShellConnection
connection_hostname = _connection.connection_hostname
connection_login = _connection.connection_login
connection_username = _connection.connection_username
local_shell_connection = _connection.local_shell_connection
open_file = _connection.open_file
register_shell_connection = _connection.register_shell_connection
shell_connection = _connection.shell_connection
ssh_shell_connection = _connection.ssh_shell_connection
is_cirros_connection = _connection.is_cirros_connection
is_local_connection = _connection.is_local_connection
put_file = _connection.put_file
put_files = _connection.put_files
get_file = _connection.get_file
make_temp_dir = _connection.make_temp_dir
make_dirs = _connection.make_dirs

ShellError = _exception.ShellError
ShellCommandFailed = _exception.ShellCommandFailed
ShellTimeoutExpired = _exception.ShellTimeoutExpired
ShellProcessTerminated = _exception.ShellProcessTerminated
ShellProcessNotTerminated = _exception.ShellProcessNotTerminated
ShellStdinClosed = _exception.ShellStdinClosed

execute = _execute.execute
execute_process = _execute.execute_process
execute_result = _execute.execute_result
ShellExecuteResult = _execute.ShellExecuteResult

HostNameError = _hostname.HostnameError
get_hostname = _hostname.get_hostname
ssh_hostname = _hostname.ssh_hostname

join_chunks = _io.join_chunks
ShellStdout = _io.ShellStdout
select_files = _io.select_files

local_execute = _local.local_execute
local_process = _local.local_process
LocalShellProcessFixture = _local.LocalShellProcessFixture
LocalExecutePathFixture = _local.LocalExecutePathFixture

ListNameserversFixture = _nameservers.ListNameserversFixture
list_nameservers = _nameservers.list_nameservers

process = _process.process
start_background_process = _process.start_background_process
check_or_start_background_process =\
    _process.check_or_start_background_process
get_user_home_dir = _path.get_user_home_dir
str_from_stream = _process.str_from_stream
ShellProcessFixture = _process.ShellProcessFixture

PsError = _ps.PsError
PsProcess = _ps.PsProcess
PsWaitTimeout = _ps.PsWaitTimeout
list_all_processes = _ps.list_all_processes
list_kernel_processes = _ps.list_kernel_processes
list_processes = _ps.list_processes
wait_for_processes = _ps.wait_for_processes

reboot_host = _reboot.reboot_host
RebootHostError = _reboot.RebootHostError
RebootHostOperation = _reboot.RebootHostOperation
RebootHostTimeoutError = _reboot.RebootHostTimeoutError
RebootHostMethod = _reboot.RebootHostMethod
crash_method = RebootHostMethod.CRASH
hard_reset_method = RebootHostMethod.HARD
soft_reset_method = RebootHostMethod.SOFT

ssh_process = _ssh.ssh_process
ssh_execute = _ssh.ssh_execute
SSHShellProcessFixture = _ssh.SSHShellProcessFixture

SystemdUnit = _systemctl.SystemdUnit
match_unit_state = _systemctl.match_unit_state
list_systemd_units = _systemctl.list_systemd_units
stop_systemd_units = _systemctl.stop_systemd_units
start_systemd_units = _systemctl.start_systemd_units
wait_for_active_systemd_units = _systemctl.wait_for_active_systemd_units
wait_for_systemd_units_state = _systemctl.wait_for_systemd_units_state

get_uptime = _uptime.get_uptime

assert_file_size = _wc.assert_file_size
get_file_size = _wc.get_file_size

CommandNotFound = _which.CommandNotFound
SkipOnCommandNotFound = _which.SkipOnCommandNotFound
find_command = _which.find_command



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko
from tobiko.shell import sh
from tobiko.openstack import stacks
from tobiko.tests.functional.openstack.stacks import test_cirros


class UbuntuServerStackTest(test_cirros.CirrosServerStackTest):
    """Tests connectivity to Nova instances via floating IPs"""

    #: Stack of resources with a server attached to a floating IP
    stack = tobiko.required_fixture(stacks.UbuntuServerStackFixture)

    nameservers_filenames = ('/run/systemd/resolve/resolv.conf',)

    def test_python(self):
        python_version = sh.execute(['python3', '--version'],
                                    ssh_client=self.stack.ssh_client).stdout
        self.assertTrue(python_version.startswith('Python 3.'),
                        python_version)

    def test_ncat_command(self):
        output = sh.execute('ncat --version',
                            ssh_client=self.stack.ssh_client).stderr.strip()
        self.assertIn('Ncat: Version', output)
        self.assertIn('https://nmap.org/ncat', output)


class UbuntuMinimalServerStackTest(UbuntuServerStackTest):

    #: Stack of resources with a server attached to a floating IP
    stack = tobiko.required_fixture(
        stacks.UbuntuMinimalServerStackFixture)

    def test_ping_fixed_ipv4(self):
        tobiko.skip_test("ping not installed on image")

    def test_ping_fixed_ipv6(self):
        tobiko.skip_test("ping not installed on image")

    def test_ncat_command(self):
        tobiko.skip_test("ncat not installed on image")



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import json
import typing

import testtools
from oslo_log import log

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack import octavia
from tobiko.openstack import stacks
from tobiko.shell import sh
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


@keystone.skip_if_missing_service(name='octavia')
class OctaviaBasicTrafficScenarioTest(testtools.TestCase):
    """Octavia traffic scenario test.

    Create a load balancer with 2 members that run a server application,
    Create a client that is connected to the load balancer VIP port,
    Generate network traffic from the client to the load balanacer.
    """
    loadbalancer_stack = tobiko.required_fixture(
        stacks.AmphoraIPv4LoadBalancerStack)

    listener_stack = tobiko.required_fixture(
        stacks.HttpRoundRobinAmphoraIpv4Listener)

    def setUp(self):
        # pylint: disable=no-member
        super(OctaviaBasicTrafficScenarioTest, self).setUp()

        # Wait for Octavia objects to be active
        LOG.info('Waiting for member '
                 f'{self.listener_stack.server_stack.stack_name} and '
                 f'for member '
                 f'{self.listener_stack.other_server_stack.stack_name} '
                 f'to be created...')
        self.listener_stack.wait_for_active_members()

        self.loadbalancer_stack.wait_for_octavia_service()

        self.listener_stack.wait_for_members_to_be_reachable()

    def test_round_robin_traffic(self):
        # For 5 minutes we ignore specific exceptions as we know
        # that Octavia resources are being provisioned
        for attempt in tobiko.retry(timeout=300.):
            try:
                octavia.check_members_balanced(
                    pool_id=self.listener_stack.pool_id,
                    ip_address=self.loadbalancer_stack.floating_ip_address,
                    lb_algorithm=self.listener_stack.lb_algorithm,
                    protocol=self.listener_stack.lb_protocol,
                    port=self.listener_stack.lb_port)
                break
            except (octavia.RoundRobinException,
                    octavia.TrafficTimeoutError,
                    sh.ShellCommandFailed):
                LOG.exception(f"Traffic didn't reach all members after "
                              f"#{attempt.number} attempts and "
                              f"{attempt.elapsed_time} seconds")
                if attempt.is_last:
                    raise


@neutron.skip_unless_is_ovn()
@keystone.skip_if_missing_service(name='octavia')
class OctaviaOVNProviderTrafficTest(testtools.TestCase):
    """Octavia OVN provider traffic test.

    Create an OVN provider load balancer with 2 members that run a server
    application,
    Create a client that is connected to the load balancer VIP port via FIP,
    Generate network traffic from the client to the load balanacer via ssh.
    """
    loadbalancer_stack = tobiko.required_fixture(
        stacks.OVNIPv4LoadBalancerStack)

    listener_stack = tobiko.required_fixture(
        stacks.TcpSourceIpPortOvnIpv4Listener)

    def setUp(self):
        # pylint: disable=no-member
        super(OctaviaOVNProviderTrafficTest, self).setUp()

        # Wait for Octavia objects to be active
        LOG.info(f'Waiting for member {self.listener_stack.member_id} and '
                 f'for member {self.listener_stack.other_member_id} '
                 f'to be created...')
        self.listener_stack.wait_for_active_members()

        self.loadbalancer_stack.wait_for_octavia_service()

    def test_ssh_traffic(self):
        """SSH every member server to get its hostname using a load balancer
        """
        username: typing.Optional[str] = None
        password: typing.Optional[str] = None
        missing_replies = set()

        for member_server in [self.listener_stack.server_stack,
                              self.listener_stack.other_server_stack]:
            ssh_client = member_server.ssh_client
            hostname = sh.get_hostname(ssh_client=ssh_client)
            missing_replies.add(hostname)
            if username is None:
                username = member_server.username
            else:
                self.assertEqual(username,
                                 member_server.username,
                                 "Not all member servers have the same "
                                 "username to login with")
            if password is None:
                password = member_server.password
            else:
                self.assertEqual(password, member_server.password,
                                 "Not all member servers have the same "
                                 "password to login with")

        # Get SSH client to the load balancer virtual IP
        ssh_client = ssh.ssh_client(
            host=self.loadbalancer_stack.floating_ip_address,
            port=self.listener_stack.lb_port,
            username=username,
            password=password)

        replies = []
        for attempt in tobiko.retry(timeout=120.):
            LOG.debug(f"SSH to member server by using the load balancer "
                      f"(login='{ssh_client.login}', attempt={attempt})...")

            with ssh_client:  # disconnect after every loop
                hostname = sh.ssh_hostname(ssh_client=ssh_client)
            try:
                missing_replies.remove(hostname)
            except KeyError:
                self.assertIn(hostname, replies,
                              f"Unexpected hostname reached: {hostname}")
            replies.append(hostname)
            if missing_replies:
                LOG.debug('Reached member server(s):\n'
                          f'{pretty_replies(replies)}')
                if attempt.is_last:
                    self.fail('Unreached member server(s): {missing_replies}')
                else:
                    LOG.debug('Waiting for reaching remaining server(s)... '
                              f'{missing_replies}')
            else:
                LOG.debug('All member servers reached:\n'
                          f'{pretty_replies(replies)}')
                break
        else:
            raise RuntimeError('Broken retry loop')


def pretty_replies(replies: typing.Iterable[str]):
    return json.dumps(collections.Counter(replies),
                      indent=4,
                      sort_keys=True)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os
import typing

import testtools

from tobiko import run


class DiscoverTestIdsTest(testtools.TestCase):

    def test_discover_test_ids(self,
                               test_files: typing.List[str] = None):
        if test_files is None:
            test_files = [__file__]
        test_ids = run.discover_test_ids(test_files=test_files)
        self.assertIn(self.id(), test_ids)

    def test_forked_discover_test_ids(self,
                                      test_files: typing.List[str] = None):
        if test_files is None:
            test_files = [__file__]
        test_ids = run.forked_discover_test_ids(test_files=test_files)
        self.assertIn(self.id(), test_ids)

    def test_find_test_ids(self,
                           test_path: typing.List[str] = None,
                           forked=False):
        if test_path is None:
            test_path = [__file__]
        test_ids = run.find_test_ids(test_path=test_path, forked=forked)
        self.assertIn(self.id(), test_ids)

    def test_find_test_ids_with_test_dir(self):
        self.test_find_test_ids(test_path=[os.path.dirname(__file__)])

    def test_find_test_ids_with_forked(self):
        self.test_find_test_ids(forked=True)



# Copyright (c) 2023 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from collections import abc
import typing

from oslo_log import log

import tobiko
from tobiko.openstack.neutron import _client


LOG = log.getLogger(__name__)
DEFAULT_SG_NAME = "default"
STATEFUL_OVN_ACTION = "allow-related"
STATELESS_OVN_ACTION = "allow-stateless"

SecurityGroupType = typing.Dict[str, typing.Any]
SecurityGroupIdType = typing.Union[str, SecurityGroupType]
SecurityGroupRuleType = typing.Dict[str, typing.Any]
SecurityGroupRuleIdType = typing.Union[str, SecurityGroupRuleType]


def list_security_groups(client=None, **params) \
        -> tobiko.Selection[SecurityGroupType]:
    security_groups = _client.neutron_client(client).list_security_groups(
        **params)
    if isinstance(security_groups, abc.Mapping):
        security_groups = security_groups['security_groups']
    return tobiko.Selection[SecurityGroupType](security_groups)


def get_default_security_group(project_id, client=None, **list_params) \
        -> SecurityGroupType:
    list_params["project_id"] = project_id
    list_params["name"] = DEFAULT_SG_NAME
    sgs = list_security_groups(
        client=client, **list_params)
    return sgs.unique


def create_security_group(client=None, add_cleanup=True,
                          **params) -> SecurityGroupType:
    sg = _client.neutron_client(client).create_security_group(
        body={'security_group': params}
    )['security_group']
    if add_cleanup:
        tobiko.add_cleanup(delete_security_group,
                           sg_id=sg['id'],
                           should_exists=False,
                           client=client)

    return sg


def delete_security_group(sg_id: SecurityGroupIdType,
                          should_exists: bool = False,
                          client: _client.NeutronClientType = None):
    try:
        _client.neutron_client(client).delete_security_group(sg_id)
    except _client.NotFound:
        if should_exists:
            raise


def create_security_group_rule(security_group_id, add_cleanup=True,
                               client=None, **rule) \
        -> SecurityGroupRuleType:
    rule['security_group_id'] = security_group_id
    sg_rule = _client.neutron_client(client).create_security_group_rule(
        body={'security_group_rule': rule}
    )['security_group_rule']

    if add_cleanup:
        tobiko.add_cleanup(delete_security_group_rule,
                           rule_id=sg_rule['id'],
                           should_exists=False,
                           client=client)

    return sg_rule


def delete_security_group_rule(rule_id: SecurityGroupRuleIdType,
                               should_exists: bool = False,
                               client: _client.NeutronClientType = None):
    try:
        _client.neutron_client(client).delete_security_group_rule(rule_id)
    except _client.NotFound:
        if should_exists:
            raise



# Copyright 2021 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import ironicclient
import ironicclient.v1.client
from oslo_log import log

import tobiko
from tobiko.openstack import _client


LOG = log.getLogger(__name__)

CLIENT_CLASSES = (ironicclient.v1.client.Client,)
IronicClient = typing.Union[ironicclient.v1.client.Client]


class IronicClientFixture(_client.OpenstackClientFixture):

    def init_client(self, session) -> IronicClient:
        return ironicclient.client.get_client(1, session=session)


class IronicClientManager(_client.OpenstackClientManager):

    def create_client(self, session) -> IronicClientFixture:
        return IronicClientFixture(session=session)


CLIENTS = IronicClientManager()

IronicClientType = typing.Union[IronicClient,
                                IronicClientFixture,
                                typing.Type[IronicClientFixture],
                                None]


def ironic_client(obj: IronicClientType) -> IronicClient:
    if obj is None:
        return get_ironic_client()

    if isinstance(obj, CLIENT_CLASSES):
        return obj

    fixture = tobiko.setup_fixture(obj)
    if isinstance(fixture, IronicClientFixture):
        assert fixture.client is not None
        return fixture.client

    message = f"Object '{obj}' is not an IronicClientFixture instance"
    raise TypeError(message)


def get_ironic_client(session=None, shared=True, init_client=None,
                      manager=None) -> IronicClient:
    manager = manager or CLIENTS
    client = manager.get_client(session=session, shared=shared,
                                init_client=init_client)
    tobiko.setup_fixture(client)
    return client.client



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko


class ShellError(tobiko.TobikoException):
    pass


class ShellCommandFailed(ShellError):
    """Raised when shell command exited with non-zero status
    """
    message = ("command '{command}' failed (exit status is {exit_status});\n"
               "stdin:\n{stdin}\n"
               "stdout:\n{stdout}\n"
               "stderr:\n{stderr}")


class ShellTimeoutExpired(ShellError):
    """Raised when shell command timeouts and has been killed before exiting
    """
    message = ("command {command} timed out after {timeout} seconds;\n"
               "stdin:\n{stdin}\n"
               "stdout:\n{stdout}\n"
               "stderr:\n{stderr}")


class ShellProcessTerminated(ShellError):
    message = ("command '{command}' terminated (exit status is {exit_status})"
               ";\n"
               "stdin:\n{stdin}\n"
               "stdout:\n{stdout}\n"
               "stderr:\n{stderr}")


class ShellProcessNotTerminated(ShellError):
    message = ("command '{command}' not terminated (time left is {time_left})"
               ";\n"
               "stdin:\n{stdin}\n"
               "stdout:\n{stdout}\n"
               "stderr:\n{stderr}")


class ShellStdinClosed(ShellError):
    message = ("command {command}: STDIN stream closed;\n"
               "stdin:\n{stdin}\n"
               "stdout:\n{stdout}\n"
               "stderr:\n{stderr}")



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os.path
import typing

import netaddr
import pytest
import testtools

import tobiko
from tobiko.shell import curl
from tobiko.shell import sh
from tobiko.shell import ssh
from tobiko.openstack import keystone
from tobiko.openstack import stacks


@pytest.mark.flaky(reruns=2, reruns_delay=60)
@keystone.skip_unless_has_keystone_credentials()
class CurlExecuteTest(testtools.TestCase):

    stack = tobiko.required_fixture(stacks.UbuntuServerStackFixture)

    def test_execute_curl(
            self,
            ip_address: typing.Optional[netaddr.IPAddress] = None,
            ssh_client: typing.Optional[ssh.SSHClientFixture] = None):
        if ip_address is None:
            # Use the floating IP
            ip_address = self.stack.ip_address
        http_port = self.stack.http_port
        result = curl.execute_curl(scheme='http',
                                   hostname=ip_address,
                                   port=http_port,
                                   path='/id',
                                   ssh_client=ssh_client,
                                   connect_timeout=10.,
                                   retry_count=30,
                                   retry_timeout=300.,
                                   retry_interval=10.).strip()
        self.assertEqual(self.stack.server_name, result)

    def test_execute_curl_ipv4(self):
        self.test_execute_curl(ip_address=self.get_fixed_ip(ip_version=4),
                               ssh_client=self.stack.ssh_client)

    def test_execute_curl_ipv6(self):
        self.test_execute_curl(ip_address=self.get_fixed_ip(ip_version=6),
                               ssh_client=self.stack.ssh_client)

    def get_fixed_ip(self, ip_version):
        for fixed_ip in self.stack.fixed_ips:
            ip_address = netaddr.IPAddress(fixed_ip['ip_address'])
            if ip_version == ip_address.version:
                return ip_address
        self.skipTest(f"Server {self.stack.server_id} has any "
                      f"IPv{ip_version} address.")


class CurlTest(testtools.TestCase):

    default_url = 'https://bootstrap.pypa.io/get-pip.py'

    def test_get_url_header(self,
                            url: str = None,
                            location: bool = None,
                            ssh_client: ssh.SSHClientType = None):
        if url is None:
            url = self.default_url
        header = curl.get_url_header(url=url,
                                     location=location,
                                     ssh_client=ssh_client)
        self.assertIsInstance(header, curl.CurlHeader)
        self.assertIn('date', header)
        return header

    def test_get_url_header_with_location(self):
        self.test_get_url_header(location=True)

    def test_get_url_header_with_no_location(self):
        self.test_get_url_header(location=True)

    def test_download_file(self,
                           url: str = None,
                           cached: bool = False,
                           download_dir: str = None,
                           **params) \
            -> curl.CurlProcessFixture:
        if url is None:
            url = tobiko.get_fixture(stacks.CirrosImageFixture).image_url
        if download_dir is None:
            download_dir = sh.make_temp_dir()
        header = self.test_get_url_header(url=url, location=True)
        process = curl.download_file(url=url,
                                     cached=cached,
                                     download_dir=download_dir,
                                     **params)
        try:
            result = process.wait()
        except RuntimeError:
            if not cached:
                raise
        else:
            self.assertIsInstance(result, sh.ShellExecuteResult)
            self.assertEqual(0, result.exit_status)

        if process.file_name is not None:
            self.assertEqual(header.content_length,
                             sh.get_file_size(process.file_name))
        return process

    def test_download_file_with_cached(self):
        process_1 = self.test_download_file(cached=True)
        self.assertTrue(process_1.executed)
        download_dir = os.path.dirname(process_1.file_name)
        process_2 = self.test_download_file(cached=True,
                                            download_dir=download_dir)
        self.assertFalse(process_2.executed)

    def test_assert_downloaded_file(self):
        process = self.test_download_file(cached=True)
        curl.assert_downloaded_file(
            file_name=process.file_name,
            headers_file_name=process.headers_file_name,
            ssh_client=process.ssh_client,
            sudo=process.sudo)



"""Models for manipulating images in/to/from storage."""

from __future__ import absolute_import

import collections
import copy
import json
import logging

from . import ConfigDict, flatten, fold_keys
from .containers import Container


class Image(collections.UserDict):
    """Model for an Image."""

    def __init__(self, client, id, data):
        """Construct Image Model."""
        # pylint: disable=redefined-builtin

        super().__init__(data)
        for k, v in data.items():
            setattr(self, k, v)

        self._id = id
        self._client = client

        assert self._id == data['id'],\
            'Requested image id({}) does not match store id({})'.format(
                self._id, data['id']
            )

    @staticmethod
    def _split_token(values=None, sep='='):
        if not values:
            return {}
        return {k: v1 for k, v1 in (v0.split(sep, 1) for v0 in values)}

    def create(self, *_args, **kwargs):
        """Create container from image.

        Pulls defaults from image.inspect()
        """
        details = self.inspect()

        config = ConfigDict(image_id=self._id, **kwargs)
        config['command'] = details.config.get('cmd')
        config['env'] = self._split_token(details.config.get('env'))
        config['image'] = copy.deepcopy(details.repotags[0])
        config['labels'] = copy.deepcopy(details.labels)
        # TODO: Are these settings still required?
        config['net_mode'] = 'bridge'
        config['network'] = 'bridge'
        config['args'] = flatten([config['image'], config['command']])

        logging.debug('Image %s: create config: %s', self._id, config)
        with self._client() as podman:
            id_ = podman.CreateContainer(config)['container']
            cntr = podman.GetContainer(id_)
        return Container(self._client, id_, cntr['container'])

    container = create

    def export(self, dest, compressed=False):
        """Write image to dest, return id on success."""
        with self._client() as podman:
            results = podman.ExportImage(self._id, dest, compressed)
        return results['image']

    def history(self):
        """Retrieve image history."""
        with self._client() as podman:
            for r in podman.HistoryImage(self._id)['history']:
                yield collections.namedtuple('HistoryDetail', r.keys())(**r)

    def inspect(self):
        """Retrieve details about image."""
        with self._client() as podman:
            results = podman.InspectImage(self._id)
        obj = json.loads(results['image'], object_hook=fold_keys())
        return collections.namedtuple('ImageInspect', obj.keys())(**obj)

    def push(self,
             target,
             compress=False,
             manifest_format="",
             remove_signatures=False,
             sign_by=""):
        """Copy image to target, return id on success."""
        with self._client() as podman:
            results = podman.PushImage(self._id, target, compress,
                                       manifest_format, remove_signatures,
                                       sign_by)
        return results['reply']['id']

    def remove(self, force=False):
        """Delete image, return id on success.

        force=True, stop any running containers using image.
        """
        with self._client() as podman:
            results = podman.RemoveImage(self._id, force)
        return results['image']

    def tag(self, tag):
        """Tag image."""
        with self._client() as podman:
            results = podman.TagImage(self._id, tag)
        return results['image']


class Images():
    """Model for Images collection."""

    def __init__(self, client):
        """Construct model for Images collection."""
        self._client = client

    def list(self):
        """List all images in the libpod image store."""
        with self._client() as podman:
            results = podman.ListImages()
        for img in results['images']:
            yield Image(self._client, img['id'], img)

    def build(self, dockerfile=None, tags=None, **kwargs):
        """Build container from image.

        See podman-build.1.md for kwargs details.
        """
        if dockerfile is None:
            raise ValueError('"dockerfile" is a required argument.')
        if not hasattr(dockerfile, '__iter__'):
            raise ValueError('"dockerfile" is required to be an iter.')

        if tags is None:
            raise ValueError('"tags" is a required argument.')
        if not hasattr(tags, '__iter__'):
            raise ValueError('"tags" is required to be an iter.')

        config = ConfigDict(dockerfile=dockerfile, tags=tags, **kwargs)
        with self._client() as podman:
            result = podman.BuildImage(config)
        return self.get(result['image']['id']), \
            (line for line in result['image']['logs'])

    def delete_unused(self):
        """Delete Images not associated with a container."""
        with self._client() as podman:
            results = podman.DeleteUnusedImages()
        return results['images']

    def import_image(self, source, reference, message='', changes=None):
        """Read image tarball from source and save in image store."""
        with self._client() as podman:
            results = podman.ImportImage(source, reference, message, changes)
        return results['image']

    def pull(self, source):
        """Copy image from registry to image store."""
        with self._client() as podman:
            results = podman.PullImage(source)
        return results['reply']['id']

    def search(self,
               id_,
               limit=25,
               is_official=None,
               is_automated=None,
               star_count=None):
        """Search registries for id."""
        constraints = {}

        if is_official is not None:
            constraints['is_official'] = is_official
        if is_automated is not None:
            constraints['is_automated'] = is_automated
        if star_count is not None:
            constraints['star_count'] = star_count

        with self._client() as podman:
            results = podman.SearchImages(id_, limit, constraints)
        for img in results['results']:
            yield collections.namedtuple('ImageSearch', img.keys())(**img)

    def get(self, id_):
        """Get Image from id."""
        with self._client() as podman:
            result = podman.GetImage(id_)
        return Image(self._client, result['image']['id'], result['image'])



# Copyright (c) 2022 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import os

import testtools

from tobiko import shiftstack
from tobiko import tripleo


PLAYBOOK_DIRNAME = os.path.join(os.path.dirname(__file__), 'playbooks')


@tripleo.skip_if_missing_tripleo_ansible_inventory
@shiftstack.skip_unless_has_shiftstack()
class OpenShiftTest(testtools.TestCase):

    def test_ocp_cluster(self):
        clouds_file_path = shiftstack.get_clouds_file_path()
        tripleo.run_playbook_from_undercloud(
            playbook='verify-shiftstack.yaml',
            playbook_dirname=PLAYBOOK_DIRNAME,
            playbook_files=[clouds_file_path],
            requirements_files=['requirements.yaml'],
            roles=['tests'])



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.openstack import neutron
from tobiko.tests.unit import openstack


class NeutronExtensionTest(openstack.OpenstackTest):

    extensions = ['ext-1', 'ext-2', 'ext-3']

    def setUp(self):
        super(NeutronExtensionTest, self).setUp()
        self.client = self.patch_get_neutron_client().return_value
        self.client.list_extensions.return_value = {
            'extensions': [{'alias': e} for e in self.extensions]}

    def test_get_networking_extensions(self):
        result = neutron.get_networking_extensions()
        self.assertEqual(frozenset(self.extensions), result)

    def test_missing_networking_extensions(self):
        result = neutron.missing_networking_extensions(
            'ext-2', 'ext-4', 'ext-0')
        self.assertEqual(['ext-0', 'ext-4'], result)

    def test_has_networking_extensions(self):
        self.assertTrue(neutron.has_networking_extensions(*self.extensions))
        self.assertFalse(neutron.has_networking_extensions('ext-1', 'ext-4'))

    def test_skip_if_networking_extensions_when_missing(self):
        ex = self.assertRaises(
            self.skipException,
            self._test_skip_if_networking_extensions_when_missing)
        self.assertEqual("missing networking extensions: ['ext-4']", str(ex))

    @neutron.skip_if_missing_networking_extensions('ext-1', 'ext-4')
    def _test_skip_if_networking_extensions_when_missing(self):
        self.fail('Not skipped')

    def test_skip_if_networking_extensions_when_not_missing(self):
        ex = self.assertRaises(
            self.failureException,
            self._test_skip_if_networking_extensions_when_not_missing)
        self.assertEqual('OK', str(ex))

    @neutron.skip_if_missing_networking_extensions('ext-1', 'ext-2')
    def _test_skip_if_networking_extensions_when_not_missing(self):
        self.fail('OK')



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

import tobiko
from tobiko.openstack import stacks
from tobiko.shell import ip
from tobiko.tests.functional.openstack.stacks import test_cirros


class VlanProxyServerStackTest(test_cirros.CirrosServerStackTest):

    #: Stack of resources with a server attached to a floating IP
    stack = tobiko.required_fixture(stacks.VlanProxyServerStackFixture)


class UbuntuVlanServerTest(testtools.TestCase):

    #: Stack of resources with a server attached to a floating IP
    stack = tobiko.required_fixture(stacks.UbuntuServerStackFixture)

    def test_vlan_ipv4_fixed_ip(self):
        self._test_vlan_fixed_ip(ip_version=4)

    def test_vlan_ipv6_fixed_ip(self):
        self._test_vlan_fixed_ip(ip_version=6)

    def _test_vlan_fixed_ip(self, ip_version: int):
        expected_ip = self.get_vlan_fixed_ip(ip_version=ip_version)
        for attempt in tobiko.retry(timeout=600.,
                                    interval=10.):
            try:
                actual_ip = ip.find_ip_address(
                    device=self.stack.vlan_device,
                    ip_version=ip_version,
                    ssh_client=self.stack.ssh_client,
                    scope='global',
                    unique=True)
            except tobiko.ObjectNotFound:
                attempt.check_limits()
            else:
                break
        else:
            raise RuntimeError('Broken retry loop')
        self.assertEqual(expected_ip, actual_ip)
        self.stack.assert_vlan_is_reachable(ip_version=ip_version)

    def get_vlan_fixed_ip(self, ip_version: int):
        try:
            return self.stack.find_vlan_fixed_ip(ip_version=ip_version)
        except tobiko.ObjectNotFound:
            self.skipTest(f"Server {self.stack.server_id} has any "
                          f"IPv{ip_version} address on VLAN device.")



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

from oslo_log import log
import pytest
import testtools

import tobiko
from tobiko import config
from tobiko.shell import files
from tobiko.shell import ping
from tobiko.shell import sh
from tobiko.openstack import neutron
from tobiko.openstack import stacks
from tobiko.openstack import topology


CONF = config.CONF
LOG = log.getLogger(__name__)


@pytest.mark.minimal
class FloatingIPTest(testtools.TestCase):
    """Tests connectivity via floating IPs"""

    #: Resources stack with floating IP and Nova server
    stack = tobiko.required_fixture(stacks.CirrosServerStackFixture)

    def test_ssh(self):
        """Test SSH connectivity to floating IP address"""
        hostname = sh.ssh_hostname(ssh_client=self.stack.ssh_client)
        self.assertEqual(self.stack.server_name.lower(), hostname)

    def test_ping(self):
        """Test ICMP connectivity to floating IP address"""
        ping.assert_reachable_hosts([self.stack.floating_ip_address])

    # --- test port-security extension ---------------------------------------

    @neutron.skip_if_missing_networking_extensions('port-security')
    def test_port_security_enabled_port_attribute(self):
        """Test port security enabled port attribute"""
        self.assertEqual(self.expected_port_security_enabled,
                         self.observed_port_security_enabled)

    @property
    def expected_port_security_enabled(self):
        """Expected port security enabled value"""
        return self.stack.port_security_enabled

    @property
    def observed_port_security_enabled(self):
        """Actual MTU value for internal network"""
        return self.stack.outputs.port_security_enabled

    # --- test security_group extension --------------------------------------

    @neutron.skip_if_missing_networking_extensions('security-group')
    def test_security_groups_port_attribute(self):
        """Test security groups port attribute"""
        self.assertEqual(self.expected_security_groups,
                         self.observed_security_groups)

    @property
    def expected_security_groups(self):
        """Expected port security groups"""
        return set(self.stack.security_groups)

    @property
    def observed_security_groups(self):
        """Actual port security group"""
        return set(self.stack.outputs.security_groups)

    # --- test net-mtu and net-mtu-writable extensions ------------------------

    @ping.skip_if_missing_fragment_ping_option
    @neutron.skip_if_missing_networking_extensions('net-mtu')
    def test_ping_with_net_mtu(self):
        """Test connectivity to floating IP address with MTU sized packets"""
        # Wait until it can reach remote port with maximum-sized packets
        ping.ping(self.stack.floating_ip_address,
                  until=ping.RECEIVED,
                  packet_size=self.observed_net_mtu,
                  fragmentation=False).assert_replied()

        # Verify it can't reach remote port with over-sized packets
        ping.ping(self.stack.floating_ip_address,
                  packet_size=self.observed_net_mtu + 1,
                  fragmentation=False,
                  count=5,
                  check=False).assert_not_replied()

    @property
    def observed_net_mtu(self):
        """Actual MTU value for internal network"""
        network_stack = self.stack.network_stack
        mtu = network_stack.mtu
        if network_stack.has_gateway:
            mtu = min(mtu, network_stack.gateway_stack.mtu)
        return mtu

    # --- test l3_ha extension ------------------------------------------------

    @neutron.skip_if_missing_networking_extensions('l3-ha')
    def test_l3_ha(self):
        """Test 'mtu' network attribute"""
        gateway = self.stack.network_stack.gateway_details
        self.assertEqual(self.stack.network_stack.ha,
                         gateway['ha'])


# --- Test with port security enabled -----------------------------------------

@neutron.skip_if_missing_networking_extensions('port-security',
                                               'security-group')
class FloatingIPWithPortSecurityFixture(stacks.CirrosServerStackFixture):
    """Heat stack for testing a floating IP instance with port security"""

    #: Resources stack with security group to allow ping Nova servers
    security_groups_stack = tobiko.required_fixture(
        stacks.SecurityGroupsFixture)

    @property
    def security_groups(self) -> typing.List[str]:
        """List with ICMP security group"""
        return [self.security_groups_stack.ssh_security_group_id]


@neutron.skip_if_missing_networking_extensions('port-security',
                                               'security-group')
class FloatingIPWithPortSecurityTest(FloatingIPTest):
    """Tests connectivity via floating IPs with port security"""

    #: Resources stack with floating IP and Nova server with port security
    stack = tobiko.required_fixture(FloatingIPWithPortSecurityFixture)

    def test_ping(self):
        """Test connectivity to floating IP address"""
        # Wait for server instance to get ready by logging in
        self.stack.ssh_client.connect()

        # Expect port security to throw away ICMP packages
        ping.assert_unreachable_hosts([self.stack.floating_ip_address],
                                      count=5, check=False)

    @ping.skip_if_missing_fragment_ping_option
    @neutron.skip_if_missing_networking_extensions('net-mtu')
    def test_ping_with_net_mtu(self):
        """Test connectivity to floating IP address"""
        # Wait for server instance to get ready by logging in
        tobiko.setup_fixture(self.stack.ssh_client)
        self.stack.ssh_client.connect()

        # Verify it can't reach secured port with maximum-sized packets
        ping.ping(self.stack.floating_ip_address,
                  packet_size=self.observed_net_mtu,
                  fragmentation=False,
                  count=5,
                  check=False).assert_not_replied()

        # Verify it can't reach secured port with over-sized packets
        ping.ping(self.stack.floating_ip_address,
                  packet_size=self.observed_net_mtu + 1,
                  fragmentation=False,
                  count=5,
                  check=False).assert_not_replied()


# --- Test with ICMP security group -------------------------------------------

class FloatingIPWithICMPSecurityGroupFixture(
        FloatingIPWithPortSecurityFixture):
    """Heat stack for testing a floating IP instance with security groups"""

    @property
    def security_groups(self):
        """List with ICMP security group"""
        return [self.security_groups_stack.ssh_security_group_id,
                self.security_groups_stack.icmp_security_group_id]


@neutron.skip_if_missing_networking_extensions('port-security',
                                               'security-group')
class FloatingIPWithICMPSecurityGroupTest(FloatingIPTest):
    """Tests connectivity via floating IP with security ICMP security group"""
    #: Resources stack with floating IP and Nova server to ping
    stack = tobiko.required_fixture(
        FloatingIPWithICMPSecurityGroupFixture)


# --- Test net-mtu-write extension --------------------------------------------

@neutron.skip_if_missing_networking_extensions('net-mtu-writable')
class FloatingIPWithNetMtuWritableFixture(stacks.CirrosServerStackFixture):
    """Heat stack for testing floating IP with a custom MTU network value"""

    #: Heat stack for creating internal network with custom MTU value
    network_stack = tobiko.required_fixture(
        stacks.NetworkWithNetMtuWriteStackFixture)


@neutron.skip_if_missing_networking_extensions('net-mtu-writable')
class FloatingIpWithMtuWritableTest(FloatingIPTest):
    """Tests connectivity via floating IP with a custom MTU value"""

    #: Resources stack with floating IP and Nova server
    stack = tobiko.required_fixture(FloatingIPWithNetMtuWritableFixture)

    def test_net_mtu_write(self):
        """Test 'mtu' network attribute"""
        self.assertEqual(self.expected_net_mtu,
                         self.stack.network_stack.mtu)

    @property
    def expected_net_mtu(self):
        """Expected MTU value for internal network"""
        return self.stack.network_stack.custom_mtu_size


# --- Test la-h3 extension ----------------------------------------------------

@neutron.skip_if_missing_networking_extensions('l3-ha')
@neutron.skip_if_missing_networking_agents(binary='neutron-l3-agent',
                                           count=2)
class FloatingIpWithL3HATest(FloatingIPTest):
    #: Resources stack with floating IP and Nova server
    stack = tobiko.required_fixture(stacks.L3haServerStackFixture)


@topology.skip_unless_osp_version('16.1', higher=True)
class TestFloatingIPLogging(testtools.TestCase):

    stack = tobiko.required_fixture(stacks.NetworkStackFixture)

    def setUp(self):
        super(TestFloatingIPLogging, self).setUp()
        self.port = neutron.create_port(network_id=self.stack.network_id)
        self.fip = neutron.create_floating_ip()
        log_filename = '/var/log/containers/neutron/server.log'
        self.log_digger = files.MultihostLogFileDigger(filename=log_filename,
                                                       sudo=True)
        for node in topology.list_openstack_nodes(group='controller'):
            self.log_digger.add_host(hostname=node.hostname,
                                     ssh_client=node.ssh_client)

    def test_fip_attach_log(self):
        '''Test log that FIP is attached to VM'''
        pattern = f'{self.fip["id"]}.*associated'
        self.log_digger.find_lines(pattern=pattern)
        neutron.update_floating_ip(
                self.fip['id'], **{'port_id': self.port['id']})
        new_logs = self.log_digger.find_new_lines(pattern=pattern)
        self.assertEqual(len(new_logs), 1)
        self.assertIn(self.port['id'], new_logs[0][1])
        self.assertIn(self.fip['floating_ip_address'], new_logs[0][1])

    def test_fip_detach_log(self):
        '''Test log that FIP is dettached from VM'''
        neutron.update_floating_ip(
                self.fip['id'], **{'port_id': self.port['id']})
        pattern = f'{self.fip["id"]}.*disassociated'
        self.log_digger.find_lines(pattern=pattern)
        neutron.update_floating_ip(self.fip['id'], **{'port_id': None})
        new_logs = self.log_digger.find_new_lines(pattern=pattern)
        self.assertEqual(len(new_logs), 1)
        self.assertIn(self.port['id'], new_logs[0][1])
        self.assertIn(self.fip['floating_ip_address'], new_logs[0][1])

    def test_fip_delete_detach_log(self):
        '''Test log that FIP is dettached from VM if FIP is deleted'''
        neutron.update_floating_ip(
                self.fip['id'], **{'port_id': self.port['id']})
        pattern = f'{self.fip["id"]}.*disassociated'
        self.log_digger.find_lines(pattern=pattern)
        neutron.delete_floating_ip(self.fip['id'])
        new_logs = self.log_digger.find_new_lines(pattern=pattern)
        self.assertEqual(len(new_logs), 1)
        self.assertIn(self.port['id'], new_logs[0][1])
        self.assertIn(self.fip['floating_ip_address'], new_logs[0][1])

    def test_port_delete_fip_detach_log(self):
        '''Test log that FIP is dettached from port if port is deleted'''
        neutron.update_floating_ip(
                self.fip['id'], **{'port_id': self.port['id']})
        pattern = f'{self.fip["id"]}.*disassociated'
        self.log_digger.find_lines(pattern=pattern)
        neutron.delete_port(self.port['id'])
        new_logs = self.log_digger.find_new_lines(pattern=pattern)
        self.assertEqual(len(new_logs), 1)
        self.assertIn(self.port['id'], new_logs[0][1])
        self.assertIn(self.fip['floating_ip_address'], new_logs[0][1])



# Copyright 2020 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.tripleo import _ansible
from tobiko.tripleo import _overcloud as overcloud
from tobiko.tripleo import _rhosp
from tobiko.tripleo import _topology as topology
from tobiko.tripleo import _undercloud as undercloud
from tobiko.tripleo import containers


get_tripleo_ansible_inventory = _ansible.get_tripleo_ansible_inventory
get_tripleo_ansible_inventory_file = \
    _ansible.get_tripleo_ansible_inventory_file
has_tripleo_ansible_inventory = _ansible.has_tripleo_ansible_inventory
read_tripleo_ansible_inventory = _ansible.read_tripleo_ansible_inventory
fetch_tripleo_inventary_file = _ansible.fetch_tripleo_inventary_file
skip_if_missing_tripleo_ansible_inventory = \
    _ansible.skip_if_missing_tripleo_ansible_inventory
skip_unless_undercloud_has_ansible = \
    _ansible.skip_unless_undercloud_has_ansible
run_playbook_from_undercloud = _ansible.run_playbook_from_undercloud

OvercloudKeystoneCredentialsFixture = \
    overcloud.OvercloudKeystoneCredentialsFixture
OvercloudNotFound = overcloud.OvercloudNotFound
OvercloudVersionMismatch = overcloud.OvercloudVersionMismatch
check_overcloud = overcloud.check_overcloud
find_overcloud_node = overcloud.find_overcloud_node
has_overcloud = overcloud.has_overcloud
list_overcloud_nodes = overcloud.list_overcloud_nodes
load_overcloud_rcfile = overcloud.load_overcloud_rcfile
overcloud_host_config = overcloud.overcloud_host_config
overcloud_keystone_client = overcloud.overcloud_keystone_client
overcloud_keystone_credentials = overcloud.overcloud_keystone_credentials
overcloud_keystone_session = overcloud.overcloud_keystone_session
overcloud_node_ip_address = overcloud.overcloud_node_ip_address
overcloud_ssh_client = overcloud.overcloud_ssh_client
overcloud_version = overcloud.overcloud_version
skip_if_missing_overcloud = overcloud.skip_if_missing_overcloud
skip_unless_has_overcloud = overcloud.skip_unless_has_overcloud
skip_if_ceph_rgw = containers.skip_if_ceph_rgw

get_rhosp_release = _rhosp.get_rhosp_release
get_rhosp_version = _rhosp.get_rhosp_version

TripleoTopology = topology.TripleoTopology

UndercloudKeystoneCredentialsFixture = \
    undercloud.UndercloudKeystoneCredentialsFixture
UndercloudCloudsFileKeystoneCredentialsFixture = \
    undercloud.UndercloudCloudsFileKeystoneCredentialsFixture
fetch_os_env = undercloud.fetch_os_env
load_undercloud_rcfile = undercloud.load_undercloud_rcfile
has_undercloud = undercloud.has_undercloud
skip_unless_has_undercloud = undercloud.skip_unlsess_has_undercloud
skip_if_missing_undercloud = undercloud.skip_if_missing_undercloud
undercloud_host_config = undercloud.undercloud_host_config
undercloud_keystone_client = undercloud.undercloud_keystone_client
undercloud_keystone_credentials = undercloud.undercloud_keystone_credentials
undercloud_keystone_session = undercloud.undercloud_keystone_session
undercloud_ssh_client = undercloud.undercloud_ssh_client
undercloud_version = undercloud.undercloud_version



# Copyright (c) 2022 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from collections import abc

import testtools

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import designate
from tobiko.openstack import stacks


@keystone.skip_unless_has_keystone_credentials()
@keystone.skip_if_missing_service(name='designate')
class DesignateClientTest(testtools.TestCase):

    stack = tobiko.required_fixture(stacks.DesignateZoneStackFixture)

    def test_get_designate_client(self):
        client = designate.get_designate_client()
        self.assertIsInstance(client, designate.DESIGNATE_CLIENT_CLASSES)

    def test_get_designate_zone(self):
        zone = designate.get_designate_zone(self.stack.zone_id)
        self.assertIsInstance(zone, abc.Mapping)
        self.assertEqual(zone['id'], self.stack.zone_id)
        self.assertEqual(zone['name'], self.stack.zone_name)
        self.assertEqual(zone, self.stack.zone_details)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import functools
import json
import sys
import typing

from oslo_log import log

import tobiko
from tobiko.shell import sh
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


_REQUIRED_CREDENTIALS_PARAMS = (
    'auth_url', 'username', 'password', 'project_name')


class KeystoneCredentials(typing.NamedTuple):
    auth_url: str
    username: str
    password: str
    project_name: str

    api_version: typing.Optional[int] = None
    domain_name: typing.Optional[str] = None
    user_domain_name: typing.Optional[str] = None
    project_domain_name: typing.Optional[str] = None
    project_domain_id: typing.Optional[str] = None
    cacert: typing.Optional[str] = None
    trust_id: typing.Optional[str] = None

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        # pylint: disable=no-member
        return {k: v
                for k, v in self._asdict().items()
                if v is not None}

    def to_json(self, indent=4, sort_keys=True) -> str:
        return json.dumps(self.to_dict(),
                          sort_keys=sort_keys,
                          indent=indent)

    def __repr__(self):
        params = self.to_dict()
        if 'password' in params:
            params['password'] = '***'
        params_dump = ', '.join(f"{name}={value!r}"
                                for name, value in sorted(params.items()))
        return f'keystone_credentials({params_dump})'

    def validate(self, required_params: typing.Iterable[str] = None):
        if required_params is None:
            required_params = _REQUIRED_CREDENTIALS_PARAMS
        missing_params = [p
                          for p in required_params
                          if not getattr(self, p)]
        if missing_params:
            reason = "undefined parameters: {!s}".format(
                ', '.join(missing_params))
            raise InvalidKeystoneCredentials(credentials=self, reason=reason)


class NoSuchKeystoneCredentials(tobiko.ObjectNotFound):
    message = "no such credentials. {reason}"


class KeystoneCredentialsFixture(tobiko.SharedFixture):

    def __init__(self,
                 credentials: KeystoneCredentials = None,
                 connection: sh.ShellConnectionType = None,
                 environ: typing.Dict[str, str] = None):
        super().__init__()
        self.credentials = credentials
        self._connection = connection
        self._environ = environ

    @property
    def connection(self) -> sh.ShellConnection:
        if self._connection is None:
            self._connection = self._get_connection()
            self.addCleanup(self._cleanup_connection)
        if not isinstance(self._connection, sh.ShellConnection):
            self._connection = sh.shell_connection(self._connection)
        return self._connection

    @property
    def environ(self) -> typing.Dict[str, str]:
        if self._environ is None:
            environ = self._get_environ()
            self._environ = {
                name: value
                for name, value in environ.items()
                if name.startswith('OS_')}
            self.addCleanup(self._cleanup_environ)
        return self._environ

    @property
    def login(self):
        return self.connection.login

    def setup_fixture(self):
        self.setup_credentials()
        assert self.credentials is not None
        self.credentials.validate()

    def setup_credentials(self) -> KeystoneCredentials:
        if self.credentials is None:
            LOG.debug('Getting credentials...\n'
                      f"  login: {self.login}\n"
                      f"  fixture: {tobiko.get_fixture_name(self)}\n")
            self.credentials = self._get_credentials()
            assert self.credentials is not None
            credentials_dump = json.dumps(self.credentials.to_dict(),
                                          sort_keys=True,
                                          indent=4)
            LOG.debug('Got credentials:\n'
                      f"  login: {self.login}\n"
                      f"  fixture: {tobiko.get_fixture_name(self)}\n"
                      "  credentials:\n"
                      f"{credentials_dump}\n")
            self.addCleanup(self._cleanup_credentials)
        return self.credentials

    def _get_credentials(self) -> KeystoneCredentials:
        raise NoSuchKeystoneCredentials(
            reason=f"[{self.fixture_name}] credentials not assigned")

    def _cleanup_credentials(self):
        self.credentials = None

    def _get_connection(self) -> sh.ShellConnectionType:
        return sh.local_shell_connection()

    def _cleanup_connection(self):
        self._connection = None

    def _get_environ(self) -> typing.Dict[str, str]:
        return self.connection.get_environ()

    def _cleanup_environ(self):
        self._environ = None

    def __repr__(self) -> str:
        return f'<{self.fixture_name} {self.login}>'


KeystoneCredentialsType = typing.Union[
    KeystoneCredentials,
    KeystoneCredentialsFixture,
    typing.Type[KeystoneCredentialsFixture]]


def keystone_credentials(credentials: KeystoneCredentialsType = None,
                         **params) \
        -> KeystoneCredentials:
    if credentials is None:
        if params:
            credentials = KeystoneCredentials(**params)
            params = {}
        else:
            credentials = default_keystone_credentials()
    assert credentials is not None

    if tobiko.is_fixture(credentials):
        credentials = tobiko.get_fixture(credentials)
        if isinstance(credentials, KeystoneCredentialsFixture):
            credentials = tobiko.setup_fixture(credentials).credentials
    assert isinstance(credentials, KeystoneCredentials)
    if params:
        params = credentials.to_dict()
        params.update(**params)
        credentials = KeystoneCredentials(**params)
    return credentials


def default_keystone_credentials() -> KeystoneCredentials:
    return tobiko.setup_fixture(
        DelegateKeystoneCredentialsFixture).credentials


class InvalidKeystoneCredentials(tobiko.TobikoException):
    message = "invalid Keystone credentials; {reason!s}; {credentials!r}"


class EnvironKeystoneCredentialsFixture(KeystoneCredentialsFixture):

    def _get_credentials(self) -> KeystoneCredentials:
        auth_url = self.get_env('OS_AUTH_URL')
        if not auth_url:
            raise NoSuchKeystoneCredentials(
                reason=(f"[{self.fixture_name}] OS_AUTH_URL environment "
                        f"variable is {auth_url!r}"))

        api_version = (
            self.get_int_env('OS_IDENTITY_API_VERSION') or
            api_version_from_url(auth_url))
        username = (
            self.get_env('OS_USERNAME') or
            self.get_env('OS_USER_ID'))
        password = self.get_env('OS_PASSWORD')
        project_name = (
            self.get_env('OS_PROJECT_NAME') or
            self.get_env('OS_TENANT_NAME') or
            self.get_env('OS_PROJECT_ID') or
            self.get_env('OS_TENANT_ID'))
        if api_version == 2:
            return keystone_credentials(
                api_version=api_version,
                auth_url=auth_url,
                username=username,
                password=password,
                project_name=project_name)
        else:
            domain_name = (
                self.get_env('OS_DOMAIN_NAME') or
                self.get_env('OS_DOMAIN_ID'))
            user_domain_name = (
                self.get_env('OS_USER_DOMAIN_NAME') or
                self.get_env('OS_USER_DOMAIN_ID'))
            project_domain_name = (
                self.get_env('OS_PROJECT_DOMAIN_NAME'))
            project_domain_id = (
                self.get_env('OS_PROJECT_DOMAIN_ID'))
            trust_id = self.get_env('OS_TRUST_ID')
            return keystone_credentials(
                api_version=api_version,
                auth_url=auth_url,
                username=username,
                password=password,
                project_name=project_name,
                domain_name=domain_name,
                user_domain_name=user_domain_name,
                project_domain_name=project_domain_name,
                project_domain_id=project_domain_id,
                trust_id=trust_id)

    def get_env(self, name) -> typing.Optional[str]:
        return self.environ.get(name)

    def get_int_env(self, name) -> typing.Optional[int]:
        value = self.get_env(name=name)
        if value is None:
            return None
        else:
            return int(value)


@functools.lru_cache()
def has_keystone_credentials(obj: KeystoneCredentialsType = None) -> bool:
    try:
        credentials = keystone_credentials(obj)
    except NoSuchKeystoneCredentials:
        LOG.debug('Openstack Keystone credentials not found', exc_info=True)
        return False
    else:
        LOG.debug(f'Openstack Keystone credentials found: {credentials!r}')
        return True


def skip_unless_has_keystone_credentials(*args, **kwargs):
    return tobiko.skip_unless('Missing Keystone credentials',
                              has_keystone_credentials, *args, **kwargs)


class ConfigKeystoneCredentialsFixture(KeystoneCredentialsFixture):

    def _get_credentials(self) -> KeystoneCredentials:
        conf = tobiko.tobiko_config().keystone
        auth_url = conf.auth_url
        if not auth_url:
            raise NoSuchKeystoneCredentials(
                reason="'auth_url' option not defined in 'keystone' section "
                       "of 'tobiko.conf' file")

        api_version = (conf.api_version or
                       api_version_from_url(auth_url))
        if api_version == 2:
            return keystone_credentials(
                api_version=api_version,
                auth_url=auth_url,
                username=conf.username,
                password=conf.password,
                project_name=conf.project_name)
        else:
            return keystone_credentials(
                api_version=api_version,
                auth_url=auth_url,
                username=conf.username,
                password=conf.password,
                project_name=conf.project_name,
                domain_name=conf.domain_name,
                user_domain_name=conf.user_domain_name,
                project_domain_name=conf.project_domain_name,
                project_domain_id=conf.project_domain_id,
                trust_id=conf.trust_id)


class DelegateKeystoneCredentialsFixture(KeystoneCredentialsFixture):

    def __init__(self,
                 delegates: typing.Iterable[KeystoneCredentialsFixture] = None,
                 credentials: KeystoneCredentials = None,
                 connection: sh.ShellConnectionType = None,
                 environ: typing.Dict[str, str] = None):
        super().__init__(credentials=credentials,
                         connection=connection,
                         environ=environ)
        if delegates is not None:
            delegates = list(delegates)
        self._delegates = delegates

    @property
    def delegates(self) -> typing.List[KeystoneCredentialsFixture]:
        if self._delegates is None:
            self._delegates = self._get_delegates()
        return self._delegates

    @staticmethod
    def _get_delegates() -> typing.List[KeystoneCredentialsFixture]:
        from tobiko.openstack.keystone import _clouds_file

        delegates: typing.List[KeystoneCredentialsFixture] = []

        keystone_conf = tobiko.tobiko_config().keystone
        hosts = keystone_conf.clouds_file_hosts
        if hosts:
            for host in hosts:
                delegates.append(
                    _clouds_file.CloudsFileKeystoneCredentialsFixture(
                        connection=host))

        proxy_client = ssh.ssh_proxy_client()
        if proxy_client is not None:
            delegates.append(
                _clouds_file.CloudsFileKeystoneCredentialsFixture(
                    connection=proxy_client))

        delegates.append(
            tobiko.get_fixture(EnvironKeystoneCredentialsFixture))
        delegates.append(
            tobiko.get_fixture(ConfigKeystoneCredentialsFixture))
        return delegates

    def _get_credentials(self) -> KeystoneCredentials:
        for delegate in self.delegates:
            try:
                return tobiko.setup_fixture(delegate).credentials
            except NoSuchKeystoneCredentials as ex:
                LOG.debug(f'Got no credentials from {delegate!r}:\n'
                          f'    {ex}\n')
        raise NoSuchKeystoneCredentials(
            reason=f'[{self.fixture_name}] no credentials from '
                   f'delegates: {self.delegates!r}')


def register_default_keystone_credentials(
        credentials: KeystoneCredentialsFixture,
        delegate: DelegateKeystoneCredentialsFixture = None,
        position: int = None):
    if delegate is None:
        delegate = tobiko.get_fixture(DelegateKeystoneCredentialsFixture)
    tobiko.check_valid_type(credentials, KeystoneCredentialsFixture)
    if position is None:
        delegate.delegates.append(credentials)
    else:
        delegate.delegates.insert(position, credentials)


def api_version_from_url(auth_url) -> typing.Optional[int]:
    if auth_url.endswith('/v2.0'):
        LOG.debug('Got Keystone API version 2 from auth_url: %r', auth_url)
        return 2
    elif auth_url.endswith('/v3'):
        LOG.debug('Got Keystone API version 3 from auth_url: %r', auth_url)
        return 3
    else:
        LOG.warning('Unable to get Keystone API version from auth_url:  %r',
                    auth_url)
        return None


def print_credentials(credentials: KeystoneCredentialsType = None):
    credentials = keystone_credentials(credentials)
    tobiko.dump_yaml(dict(credentials.to_dict()),
                     sys.stdout,
                     indent=4,
                     sort_keys=True)



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.shell.tcpdump import _parameters


def get_tcpdump_command(parameters: _parameters.TcpdumpParameters):
    interface = TcpdumpInterface()
    return interface.get_tcpdump_command(parameters)


class TcpdumpInterface:

    def get_tcpdump_command(
            self, parameters: _parameters.TcpdumpParameters) -> str:
        command = 'tcpdump -s0 -Un'
        if parameters.capture_timeout is not None:
            command = f'timeout {parameters.capture_timeout} ' + command
        options = self.get_tcpdump_options(parameters=parameters)
        return command + ' ' + options

    def get_tcpdump_options(
            self,
            parameters: _parameters.TcpdumpParameters) -> str:
        options = f'-w {parameters.capture_file}'
        if parameters.interface is not None:
            options += f' -i {parameters.interface}'
        else:
            options += ' -i any'
        if parameters.capture_filter is not None:
            options += f' {parameters.capture_filter}'
        return options



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import pytest
from oslo_log import log
import testtools

import tobiko
from tobiko.openstack import neutron
from tobiko.openstack import topology
from tobiko.openstack import tests
from tobiko.tests.faults.ha import cloud_disruptions
from tobiko.tripleo import pacemaker
from tobiko.tripleo import processes
from tobiko.tripleo import containers
from tobiko.tripleo import nova
from tobiko.tripleo import overcloud
from tobiko.tripleo import undercloud
from tobiko.tripleo import validations


LOG = log.getLogger(__name__)


def overcloud_health_checks(passive_checks_only=False,
                            skip_mac_table_size_test=False):
    # this method will be changed in future commit
    check_pacemaker_resources_health()
    check_overcloud_processes_health()
    nova.check_nova_services_health()
    tests.test_neutron_agents_are_alive()
    if not passive_checks_only:
        # create a uniq stack
        check_vm_create()
        nova.start_all_instances()
        nova.check_computes_vms_running_via_virsh()
    containers.list_node_containers.cache_clear()
    containers.assert_all_tripleo_containers_running()
    containers.assert_equal_containers_state()
    containers.run_container_config_validations()
    tests.test_ovn_dbs_validations()
    # skip_mac_table_size_test has to be removed when BZ1695122 is resolved
    # we need it for the moment because this validation should not be performed
    # after any overcloud node is rebooted
    if not skip_mac_table_size_test:
        tests.test_ovs_bridges_mac_table_size()
    validations.run_post_deployment_validations()


# check vm create with ssh and ping checks
def check_vm_create():
    tests.test_server_creation()


# check cluster failed statuses
def check_pacemaker_resources_health():
    return pacemaker.PacemakerResourcesStatus().all_healthy


def check_overcloud_processes_health():
    procs = processes.OvercloudProcessesStatus()
    return (procs.basic_overcloud_processes_running and
            procs.ovn_overcloud_processes_validations)


class OvercloudHealthCheck(tobiko.SharedFixture):

    skips: typing.FrozenSet[str] = frozenset()

    @classmethod
    def run_before(cls, **params):
        cls.run(after=False, **params)

    @classmethod
    def run_after(cls, **params):
        cls.run(after=True, **params)

    @classmethod
    def run(cls, after: bool, **params):
        fixture = tobiko.get_fixture(cls)
        params.setdefault('passive_checks_only', False)
        # In version OSP17.0 and highier,
        # 'test_ovs_bridges_mac_table_size()' test can run.
        if topology.verify_osp_version('17.0', lower=True):
            params.setdefault('skip_mac_table_size_test', True)
        skips = frozenset(k for k, v in params.items() if v)
        if after or skips < fixture.skips:
            # Force re-check
            tobiko.cleanup_fixture(fixture)
        else:
            LOG.info("Will skip Overcloud health checks if already "
                     f"executed: {params}")
        fixture.skips = skips
        tobiko.setup_fixture(fixture)

    def setup_fixture(self):
        # run validations
        params = {name: True
                  for name in self.skips}
        LOG.info(f"Start executing Overcloud health checks: {params}.")
        overcloud_health_checks(**params)
        LOG.info(f"Overcloud health checks successfully executed: {params}.")

    def cleanup_fixture(self):
        self.skips = frozenset()


@undercloud.skip_if_missing_undercloud
class DisruptTripleoNodesTest(testtools.TestCase):
    """ HA Tests: run health check -> disruptive action -> health check
    disruptive_action: a function that runs some
    disruptive scenario on a overcloud"""

    def test_0vercloud_health_check(self):
        OvercloudHealthCheck.run_before(skip_mac_table_size_test=False)

    @nova.skip_background_vm_ping_checks
    def test_z99_hard_reboot_controllers_recovery(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.reset_all_controller_nodes()
        OvercloudHealthCheck.run_after()

    @nova.skip_background_vm_ping_checks
    def test_soft_reboot_computes_recovery(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.reset_all_compute_nodes(hard_reset=False)
        # verify VM status is updated after reboot
        nova.wait_for_all_instances_status('SHUTOFF')
        # start all VM instance
        # otherwise sidecar containers will not run after computes reboot
        nova.start_all_instances()
        OvercloudHealthCheck.run_after(passive_checks_only=True)

    # TODO(eolivare): the following test is skipped due to rhbz#1890895
    # def test_hard_reboot_computes_recovery(self):
    #     OvercloudHealthCheck.run_before()
    #     cloud_disruptions.reset_all_compute_nodes(hard_reset=True)
    #     # verify VM status is updated after reboot
    #     nova.wait_for_all_instances_status('SHUTOFF')
    #     # start all VM instance
    #     # otherwise sidecar containers will not run after computes reboot
    #     nova.start_all_instances()
    #     OvercloudHealthCheck.run_after(passive_checks_only=True)

    def test_z99_reboot_controller_main_vip(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.reset_controller_main_vip()
        OvercloudHealthCheck.run_after()

    def test_z99_reboot_controller_non_main_vip(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.reset_controllers_non_main_vip()
        OvercloudHealthCheck.run_after()

    def test_z99_crash_controller_main_vip(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.crash_controller_main_vip()
        OvercloudHealthCheck.run_after()

    @overcloud.skip_unless_kexec_tools_installed
    def test_z99_crash_controller_non_main_vip(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.crash_controllers_non_main_vip()
        OvercloudHealthCheck.run_after()

    @pacemaker.skip_if_fencing_not_deployed
    def test_network_disruptor_main_vip(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.network_disrupt_controller_main_vip()
        OvercloudHealthCheck.run_after()
        cloud_disruptions.network_undisrupt_controller_main_vip()

    # @pacemaker.skip_if_fencing_not_deployed
    # def test_network_disruptor_non_main_vip(self):
    #     OvercloudHealthCheck.run_before()
    #     cloud_disruptions.network_disrupt_controllers_non_main_vip()
    #     OvercloudHealthCheck.run_after()
    #     cloud_disruptions.network_undisrupt_controllers_non_main_vip()

    @neutron.skip_unless_is_ovn()
    @overcloud.skip_unless_ovn_using_ha
    def test_reset_ovndb_pcs_master_resource(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.reset_ovndb_pcs_master_resource()
        OvercloudHealthCheck.run_after()

    @neutron.skip_unless_is_ovn()
    @overcloud.skip_unless_ovn_using_ha
    def test_reset_ovndb_pcs_resource(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.reset_ovndb_pcs_resource()
        OvercloudHealthCheck.run_after()

    @neutron.skip_unless_is_ovn()
    @overcloud.skip_unless_ovn_using_ha
    def test_reset_ovndb_master_container(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.reset_ovndb_master_container()
        OvercloudHealthCheck.run_after()

    def test_kill_rabbitmq_service_one_controller(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.kill_rabbitmq_service()
        OvercloudHealthCheck.run_after()

    def test_kill_all_galera_services(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.kill_all_galera_services()
        OvercloudHealthCheck.run_after()

    def test_remove_all_grastate_galera(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.remove_all_grastate_galera()
        OvercloudHealthCheck.run_after()

    def test_remove_one_grastate_galera(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.remove_one_grastate_galera()
        OvercloudHealthCheck.run_after()

    def test_request_galera_sst(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.request_galera_sst()
        OvercloudHealthCheck.run_after()

    @pytest.mark.flaky(reruns=3, reruns_delay=60)
    def test_controllers_shutdown(self):
        OvercloudHealthCheck.run_before()
        cloud_disruptions.test_controllers_shutdown()
        OvercloudHealthCheck.run_after()

# [..]
# more tests to follow
# run health checks
# faults stop rabbitmq service on one controller
# run health checks again



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

from tobiko.shell import sh


class TestSystemctl(testtools.TestCase):

    def setUp(self):
        super().setUp()
        try:
            sh.execute('pgrep systemd')
        except sh.ShellCommandFailed:
            self.skipTest("systemd is not running")

    def test_list_system_services(self):
        units = sh.list_systemd_units()
        self.assertNotEqual([], units)

    def test_list_system_services_with_all(self):
        units = sh.list_systemd_units(all=True)
        self.assertNotEqual([], units)

    def test_list_system_services_with_state(self):
        units = sh.list_systemd_units(state='active')
        self.assertNotEqual([], units)
        self.assertEqual(units.with_attributes(active='active'), units)

    def test_list_system_services_with_type(self):
        units = sh.list_systemd_units(type='socket')
        self.assertNotEqual([], units)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from collections import abc
import os
import sys
import typing

from heatclient.common import template_utils

import tobiko


TEMPLATE_SUFFIX = '.yaml'

TEMPLATE_DIRS = list(sys.path)


class HeatTemplateFixture(tobiko.SharedFixture):

    template_yaml: str

    def __init__(self,
                 template: typing.Mapping[str, typing.Any] = None,
                 template_files: typing.Mapping = None):
        super(HeatTemplateFixture, self).__init__()
        self.template: typing.Dict[str, typing.Any] = {}
        if template is not None:
            self.template.update(template)
        self.template_files: typing.Dict[str, typing.Any] = {}
        if template_files is not None:
            self.template_files.update(template_files)

    def setup_fixture(self):
        self.setup_template()

    def setup_template(self):
        # Ensure main sections are dictionaries
        tobiko.check_valid_type(self.outputs, abc.Mapping)
        tobiko.check_valid_type(self.parameters, abc.Mapping)
        tobiko.check_valid_type(self.resources, abc.Mapping)
        self.template_yaml = tobiko.dump_yaml(self.template)

    @property
    def outputs(self) -> typing.Dict[str, typing.Any]:
        return dict(self.template.get('outputs', {}))

    @property
    def parameters(self) -> typing.Dict[str, typing.Any]:
        return dict(self.template.get('parameters', {}))

    @property
    def resources(self) -> typing.Dict[str, typing.Any]:
        return dict(self.template.get('resources', {}))


class HeatTemplateFileFixture(HeatTemplateFixture):

    def __init__(self,
                 template_file: str,
                 template_dirs: typing.Iterable[str] = None):
        super(HeatTemplateFileFixture, self).__init__()
        self.template_file = template_file
        if template_dirs is None:
            template_dirs = TEMPLATE_DIRS
        self.template_dirs: typing.List[str] = list(template_dirs)

    def setup_template(self):
        template_file = self.template_file
        if self.template_dirs or not os.path.isfile(template_file):
            template_dirs = self.template_dirs or TEMPLATE_DIRS
            template_file = find_heat_template_file(
                template_file=self.template_file,
                template_dirs=template_dirs)
        template_files, template = template_utils.get_template_contents(
            template_file=template_file)
        self.template = template
        self.template_files = template_files
        super(HeatTemplateFileFixture, self).setup_template()


HeatTemplateType = typing.Union[typing.Mapping[str, typing.Any],
                                HeatTemplateFixture]


def heat_template(obj: HeatTemplateType,
                  template_files: typing.Mapping = None) \
        -> HeatTemplateFixture:
    if isinstance(obj, abc.Mapping):
        template = HeatTemplateFixture(template=obj,
                                       template_files=template_files)
    else:
        template = tobiko.get_fixture(obj)

    tobiko.check_valid_type(template, HeatTemplateFixture)
    return template


def heat_template_file(template_file: str,
                       template_dirs: typing.Iterable[str] = None):
    return HeatTemplateFileFixture(template_file=template_file,
                                   template_dirs=template_dirs)


def find_heat_template_file(template_file: str,
                            template_dirs: typing.Iterable[str]):
    for template_dir in template_dirs:
        template_path = os.path.join(template_dir, template_file)
        if os.path.exists(template_path):
            return template_path

    msg = "Template file {!r} not found in directories {!r}".format(
        template_file, template_dirs)
    raise IOError(msg)
